import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from .internals import Dataset, field, register


@register(
    body_region=('Head', 'Abdominal'),
    license=None,  # FIXME: inherit licenses from the original datasets...
    link='http://medicalood.dkfz.de/web/',
    modality=('MRI', 'CT'),
    prep_data_size='405G',
    raw_data_size='120G',
    task='Out-of-distribution detection',
)
class MOOD(Dataset):
    """
    A (M)edival (O)ut-(O)f-(D)istribution analysis challenge [1]_

    This dataset contains raw brain MRI and abdominal CT images.

    Number of training samples:
    - Brain: 800 scans ( 256 x 256 x 256 )
    - Abdominal: 550 scans ( 512 x 512 x 512 )

    For each setup there are 4 toy test samples with OOD cases.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at https://www.synapse.org/#!Synapse:syn21343101/wiki/599515.

    Then, the folder with raw downloaded data should contain four zip archives with data
    (`abdom_toy.zip`, `abdom_train.zip`, `brain_toy.zip` and `brain_train.zip`).

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = MOOD(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 1358
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 512)
    >>> print(ds.pixel_label(ds.ids[0]).shape)
    # (512, 512, 512)

    References
    ----------
    .. [1] Zimmerer, Petersen, et al. "Medical Out-of-Distribution Analysis Challenge 2022."
           doi: 10.5281/zenodo.6362313 (2022).
    """

    @property
    def ids(self):
        result = set()
        # zip archives for train images:
        for archive in self.root.glob('*.zip'):
            if 'brain' in str(archive):  # define whether it is brain (MRI) or abdominal (CT)
                task = 'brain'
            else:
                task = 'abdom'

            if 'toy' in str(archive):  # fold - train or toy test
                fold = 'toy'
            else:
                fold = 'train'

            with ZipFile(archive) as zf:
                for zipinfo in zf.infolist():
                    if zipinfo.is_dir():
                        continue

                    file_stem = Path(zipinfo.filename).stem
                    if '.nii' in file_stem:
                        if fold == 'train':
                            result.add(f'mood_{task}_{fold}_{file_stem.split(".nii")[0]}')
                        # fold == 'toy'
                        else:
                            result.add(f'mood_{task}_{file_stem.split(".nii")[0]}')

        return tuple(sorted(result))

    @field
    def fold(self, i):
        """Returns fold: train or toy (test)."""
        if 'train' in i:
            return 'train'
        # if 'toy' in i:
        return 'toy'

    @field
    def task(self, i):
        """Returns task: brain (MRI) or abdominal (CT)."""
        if 'brain' in i:
            return 'brain'
        # if 'abdom' in i:
        return 'abdom'

    def _file(self, i):
        task, fold, num_id = i.split('_')[-3:]
        if fold == 'train':
            return zipfile.Path(self.root / f'{task}_{fold}.zip', f'{task}_{fold}/{num_id}.nii.gz')
        return zipfile.Path(self.root / f'{task}_{fold}.zip', f'toy/toy_{num_id}.nii.gz')

    @field
    def image(self, i):
        with open_nii_gz_file(self._file(i)) as nii_image:
            return np.asarray(nii_image.dataobj)

    @field
    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(self._file(i)) as nii_image:
            return nii_image.affine

    def spacing(self, i):
        """Returns voxel spacing along axes (x, y, z)."""
        with open_nii_gz_file(self._file(i)) as nii_image:
            return tuple(nii_image.header['pixdim'][1:4])

    @field
    def sample_label(self, i):
        """
        Returns sample-level OOD score for toy examples and None otherwise.
        0 indicates no abnormality and 1 indicates abnormal input.
        """
        file = self._file(i)
        if 'toy' in file.name:
            with (file.parent.parent / 'toy_label/sample' / f'{file.name}.txt').open('r') as nii:
                return int(nii.read())

    @field
    def pixel_label(self, i):
        """
        Returns voxel-level OOD scores for toy examples and None otherwise.
        0 indicates no abnormality and 1 indicates abnormal input.
        """
        file = self._file(i)
        if 'toy' in file.name:
            with open_nii_gz_file(file.parent.parent / 'toy_label/pixel' / file.name) as nii_image:
                return np.bool_(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})



import contextlib
import gzip
import json
import tarfile
import typing as tp
from functools import cached_property
from pathlib import Path

import nibabel as nb
import numpy as np
import pandas as pd
import pydicom
from deli import load

from .internals import Dataset, field, licenses, register


@register(
    body_region='Chest',
    license=licenses.CC_BY_40,
    link='https://ieee-dataport.org/open-access/bimcv-covid-19'
    '-large-annotated-dataset-rx-and-ct-images-covid-19-patients-0',
    modality='CT',
    prep_data_size='859G',
    raw_data_size='859G',
    task='Segmentation',
)
class BIMCVCovid19(Dataset):
    """
    BIMCV COVID-19 Dataset, CT-images only
    It includes BIMCV COVID-19 positive partition (https://arxiv.org/pdf/2006.01174.pdf)
    and negative partion
    (https://ieee-dataport.org/open-access/bimcv-covid-19-large-annotated-dataset-rx-and-ct-images-covid-19-patients-0)

    PCR tests are not used

    GitHub page: https://github.com/BIMCV-CSUSP/BIMCV-COVID-19

    Parameters
    ----------
    root : str, Path
        path to the folder containing the downloaded and parsed data.

    Notes
    -----
    Dataset has 2 partitions: bimcv-covid19-positive and bimcv-covid19-positive
    Each partition is spread over the 81 different tgz archives. The archives include metadata about
    subject, sessions, and labels. Also, there are some tgz archives for nifty images in nii.gz format

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = BIMCVCovid19(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 201
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 163)
    >>> print(ds.is_positive(ds.ids[0]))
    # True
    >>> print(ds.subject_info[80])
    # {'modality_dicom': "['CT']",
    #  'body_parts': "[['chest']]",
    #  'age': '[80]',
    #  'gender': 'M'}

    References
    ----------
    .. [1] Maria De La Iglesia Vayá, Jose Manuel Saborit, Joaquim Angel Montell, Antonio Pertusa, Aurelia
            Bustos, Miguel Cazorla, Joaquin Galant, Xavier Barber, Domingo Orozco-Beltrán, Francisco
            Garcia, Marisa Caparrós, Germán González, and Jose María Salinas. BIMCV COVID-19+: a
            large annotated dataset of RX and CT images from COVID-19 patients. arXiv:2006.01174, 2020.
    .. [2] Maria de la Iglesia Vayá, Jose Manuel Saborit-Torres, Joaquim Angel Montell Serrano,
            Elena Oliver-Garcia, Antonio Pertusa, Aurelia Bustos, Miguel Cazorla, Joaquin Galant,
            Xavier Barber, Domingo Orozco-Beltrán, Francisco García-García, Marisa Caparrós, Germán González,
            Jose María Salinas, 2021. BIMCV COVID-19-: a large annotated dataset of RX and CT images from COVID-19
            patients.
            Available at: https://dx.doi.org/10.21227/m4j2-ap59.
    """

    @property
    def ids(self):
        return tuple(sorted(self._series2metainfo))

    @field
    def session_id(self, i):
        return self._series2metainfo[i]['session_id']

    @field
    def subject_id(self, i):
        return self._series2metainfo[i]['subject_id']

    @field
    def is_positive(self, i):
        return self._series2metainfo[i]['is_positive']

    @field
    def image(self, i):
        meta = self._series2metainfo[i]
        with unpack(self._current_root(i), meta['archive_path'], meta['image_path']) as (file, unpacked):
            if unpacked:
                array = np.asarray(nb.load(file).dataobj)
            else:
                with gzip.GzipFile(fileobj=file) as nii:
                    nii = nb.FileHolder(fileobj=nii)
                    image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                    array = np.asarray(image.dataobj)

            integer = np.int16(array)
            if (array == integer).all():
                return integer

            # there are about 30 images that will visit this branch, so it's ok
            integer = np.int32(array)
            assert (array == integer).all(), np.abs(array - integer.astype(float)).max()
            return integer

    @field
    def affine(self, i):
        meta = self._series2metainfo[i]
        with unpack(self._current_root(i), meta['archive_path'], meta['image_path']) as (file, unpacked):
            if unpacked:
                return nb.load(file).affine

            with gzip.GzipFile(fileobj=file) as nii:
                nii = nb.FileHolder(fileobj=nii)
                image = nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def tags(self, i) -> dict:
        """
        dicom tags
        """
        meta = self._series2metainfo[i]
        if meta['meta_path'] is None:
            return {}

        with unpack(self._current_root(i), meta['archive_path'], meta['meta_path']) as (file, _):
            try:
                return parse_dicom_tags(load(file))
            except ValueError:
                return {}

    @cached_property
    def _labels_dataframe(self):
        with unpack(
            self._positive_root, 'covid19_posi_head.tar.gz', 'covid19_posi/derivatives/labels/labels_covid_posi.tsv'
        ) as (file, _):
            pos_dataframe = pd.read_csv(file, sep='\t', index_col='ReportID').iloc[:, 1:]

        with unpack(
            self._negative_root,
            'covid19_neg_derivative.tar.gz',
            'covid19_neg/derivatives/labels/Labels_covid_NEG_JAN21.tsv',
        ) as (file, _):
            neg_dataframe = pd.read_csv(file, sep='\t', index_col='ReportID').iloc[:, 1:]

        return pd.concat([pos_dataframe, neg_dataframe], ignore_index=False)

    @field
    def label_info(self, i) -> dict:
        """
        labelCUIS, Report, LocalizationsCUIS etc.
        """
        session_id = self._series2metainfo[i]['session_id']

        if session_id in self._labels_dataframe.index:
            return dict(self._labels_dataframe.loc[session_id])
        else:
            return {}

    @cached_property
    def _subject_df(self):
        with unpack(self._positive_root, 'covid19_posi_subjects.tar.gz', 'covid19_posi/participants.tsv') as (file, _):
            pos_data = pd.read_csv(file, sep='\t', index_col='participant')
            pos_data = pos_data[pos_data.index != 'derivatives']

        with unpack(self._negative_root, 'covid19_neg_metadata.tar.gz', 'covid19_neg/participants.tsv') as (file, _):
            neg_data = pd.read_csv(file, sep='\t', index_col='participant')
            neg_data = neg_data[neg_data.index != 'derivatives']

        return pd.concat([pos_data, neg_data])

    @field
    def subject_info(self, i) -> dict:
        """
        modality_dicom (=[CT]), body_parts(=[chest]), age, gender
        """
        subject_id = self._series2metainfo[i]['subject_id']
        if subject_id in self._subject_df.index:
            return dict(self._subject_df.loc[subject_id])
        else:
            return {}

    @field
    def age(self, i) -> int:
        """Minimum of (possibly two) available ages.
        The maximum difference between max and min age for every patient is 1 year."""
        return min(json.loads(self.subject_info(i).get('age')))

    @field
    def sex(self, i) -> str:
        return self.subject_info(i).get('gender')

    @field
    def session_info(self, i) -> dict:
        """
        study_date,	medical_evaluation
        """
        meta = self._series2metainfo[i]
        current_root = self._current_root(i)
        session_id = meta['session_id']
        subject_id = meta['subject_id']

        if meta['is_positive']:
            step_sessions_tarfile_name = 'covid19_posi_sessions_tsv.tar.gz'
        else:
            step_sessions_tarfile_name = 'covid19_neg_sessions_tsv.tar.gz'

        txt_splits = load(current_root / (step_sessions_tarfile_name + '.tar-tvf.txt')).split()
        (ses_file,) = filter(lambda x: subject_id in x, txt_splits)

        with unpack(current_root, step_sessions_tarfile_name, ses_file) as (file, _):
            sesions_dataframe = pd.read_csv(file, sep='\t', index_col='session_id')

        if session_id in sesions_dataframe.index:
            return dict(sesions_dataframe.loc[session_id])
        return {}

    @cached_property
    def _series2metainfo(self):
        """
        Main function that gathers the metadata for the whole dataset.

        Returns
        -------
        A dict with {series_id : {'session_id': ...
                                'subject_id': ...,
                                'archive_path': ...,
                                'is_positive': ...,
                                'image_path': ...,
                                'meta_path': ...}}
        """
        series2metainfo = {}
        for part in self._positive_root, self._negative_root:
            for structure in part.glob('*part*.tar.gz.tar-tvf.txt'):
                part_filename = structure.name[: -len('.tar-tvf.txt')]
                if 'pos' in part_filename:
                    is_positive = True
                else:
                    assert 'neg' in part_filename
                    is_positive = False

                text_descr_splits = load(structure).split()
                for el_desrc in text_descr_splits:
                    # only chest ct images ('.json' and '.nii.gz' files)
                    # 'don't remove png but it should not be in cases'
                    if 'chest_ct' not in el_desrc or not el_desrc.endswith(('.nii.gz', '.json')):
                        continue

                    image_path = None
                    meta_path = None
                    if el_desrc.endswith('.nii.gz'):
                        ext = '.nii.gz'
                        image_path = el_desrc
                    # elif el_desrc.endswith(".json"):
                    else:
                        assert el_desrc.endswith('.json')
                        ext = '.json'
                        meta_path = el_desrc

                    # obtaining series id
                    series_id = str(Path(el_desrc).name)[: -len(ext)]

                    # if no such series yet
                    if series_id not in series2metainfo:
                        # obtain subject_id, session_id
                        (subject_id,) = filter(lambda x: 'sub' in x, series_id.split('_'))
                        (session_id,) = filter(lambda x: 'ses' in x, series_id.split('_'))

                        series2metainfo[series_id] = {
                            'session_id': session_id,
                            'subject_id': subject_id,
                            'archive_path': part_filename,
                            'is_positive': is_positive,
                            'image_path': None,
                            'meta_path': None,
                        }

                    if image_path is not None:
                        series2metainfo[series_id]['image_path'] = image_path

                    if meta_path is not None:
                        series2metainfo[series_id]['meta_path'] = meta_path

        return series2metainfo

    @cached_property
    def _positive_root(self):
        return find_subroot(self.root, 'covid19_posi')

    @cached_property
    def _negative_root(self):
        return find_subroot(self.root, 'covid19_neg')

    def _current_root(self, i):
        return self._positive_root if self._series2metainfo[i]['is_positive'] else self._negative_root


@contextlib.contextmanager
def unpack(root: Path, archive: str, relative: str):
    unpacked = root / relative
    if unpacked.exists():
        yield unpacked, True
    else:
        # we use a buffer of 128mb to speed everything up
        with tarfile.open(root / archive, bufsize=128 * 1024**2) as part_file:
            yield part_file.extractfile(relative), False


def find_subroot(path: Path, name: str):
    """Finds a subfolder but in a bfs manner instead of dfs"""
    folders = [path]

    while folders:
        current = folders.pop(0)

        for child in current.iterdir():
            if child.name.startswith(name) and child.name.endswith('.tar-tvf.txt'):
                return current

            if child.is_dir():
                folders.append(child)

    raise FileNotFoundError(
        'No "*.tar-tvf.txt" files have been found. They are needed to gather the datasets structure.'
    )


def parse_dicom_tags(tags: tp.Dict[str, tp.Any]) -> tp.Optional[tp.Union[dict, list]]:
    if not isinstance(tags, dict):
        return tags
    if len(tags) == 1:
        if 'vr' in tags:
            return None
        return parse_dicom_tags(list(tags.values())[0])

    if set(tags.keys()) == {'Value', 'vr'}:
        value = tags['Value']
        # vr = tags["vr"]
        assert isinstance(value, list)
        result = [parse_dicom_tags(v) for v in value]
        if not result:
            return None
        if len(result) == 1:
            return result[0]
        return result

    result_: tp.Dict[str, tp.Optional[tp.Union[dict, list]]] = {}
    for tag, value in tags.items():
        try:
            keyword = pydicom.datadict.keyword_for_tag(tag)
        except ValueError:
            keyword = str(tag)
        assert isinstance(result_, dict)
        result_[keyword] = parse_dicom_tags(value)
    return result_



from typing import Sequence, Union

import numpy as np
from connectome import Transform
from imops import zoom

from ..utils import Numeric, propagate_none


class CanonicalCTOrientation(Transform):
    __inherit__ = True

    def image(image):
        return np.transpose(image, (1, 0, 2))[::-1, :, ::-1]

    def mask(mask):
        return np.transpose(mask, (1, 0, 2))[::-1, :, ::-1]

    def spacing(spacing):
        return tuple(np.array(spacing)[[1, 0, 2]].tolist())


class Rescale(Transform):
    __exclude__ = (
        'voxel_spacing',
        'affine',
    )

    _new_spacing: Union[Sequence[Numeric], Numeric]
    _order: int = 1

    def _spacing(spacing, _new_spacing):
        _new_spacing = np.broadcast_to(_new_spacing, len(spacing)).copy()
        _new_spacing[np.isnan(_new_spacing)] = np.array(spacing)[np.isnan(_new_spacing)]
        return tuple(_new_spacing.tolist())

    def _scale_factor(spacing, _spacing):
        return np.float32(spacing) / np.float32(_spacing)

    def spacing(_spacing):
        return _spacing

    def image(image, _scale_factor, _order):
        return zoom(image.astype(np.float32), _scale_factor, order=_order)

    @propagate_none
    def mask(mask, _scale_factor, _order):
        onehot = np.arange(mask.max() + 1) == mask[..., None]
        onehot = onehot.astype(mask.dtype).transpose(3, 0, 1, 2)
        out = np.array(zoom(onehot.astype(np.float32), _scale_factor, axis=(1, 2, 3)) > 0.5, dtype=mask.dtype)
        labels = out.argmax(axis=0)
        return labels



import nibabel
import numpy as np
from connectome import Output, Transform


class SpacingFromAffine(Transform):
    __inherit__ = True

    def spacing(affine):
        return nibabel.affines.voxel_sizes(affine)


class ParseAffineMatrix(Transform):
    """Splits affine matrix into separate methods for more convenient usage.

    Examples
    --------
    >>> dataset = Dataset()
    >>> dataset.voxel_spacing(id_)
    # FieldError
    >>> dataset = dataset >> ParseAffineMatrix()
    >>> dataset.voxel_spacing(id_)
    # array([1.5, 1.5, 1.5])
    """

    __inherit__ = True

    def origin(affine):
        """Constructs an origin tensor from the given affine matrix."""
        return affine[:-1, -1]

    def spacing(affine):
        """Constructs a voxel spacing tensor from the given orientation matrix."""
        return np.linalg.norm(affine[:3, :3], axis=0)

    def orientation(affine, spacing: Output):
        """Constructs an orientation matrix from the given affine matrix."""
        return np.divide(affine[:3, :3], spacing)



import gzip
import zipfile
from typing import Dict
from zipfile import ZipFile

import nibabel
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region='Abdomen',
    license=licenses.CC_BY_40,
    link='https://zenodo.org/records/13767408',
    modality='CT',
    prep_data_size='30G',
    raw_data_size='30G',
    task='Abdominal organ pathologies segmentation',
)
class CURVAS(Dataset):
    """
    Pancreas, liver and kidney cysts segmentation from multi-rater annotated data.

    The dataset was used at the MICCAI 2024 CURVAS challenge.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Download link: https://zenodo.org/records/13767408

    The `root` folder should contain the three downloaded .zip archives, namely:
    `training_set.zip`, `validation_set.zip` and `testing_set.zip`.

    Examples
    --------
    >>> # Place the downloaded folders in any folder and pass the path to the constructor:
    >>> ds = CURVAS(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 90
    >>> print(ds.image(ds.ids[5]).shape)
    # (512, 512, 1045)
    >>> print(ds.mask(ds.ids[35]).shape)
    # (512, 512, 992)

    """

    @property
    def ids(self):
        def _extract(split):
            archive = self.root / f'{split}_set.zip'
            with ZipFile(archive) as zf:
                namelist = [x for x in zf.namelist() if len(x.rstrip('/').split('/')) == 2]
                ids = [f'{x.split("/")[1]}-{split}' for x in namelist]
                return ids

        return sorted(
            [
                *_extract('training'),  # 20 Training   cases
                *_extract('validation'),  # 5  Validation cases
                *_extract('testing'),  # 65 Testing    cases
            ]
        )

    def _file(self, i, obj):
        uid, split = i.split('-')

        archive = self.root / f'{split}_set.zip'
        file = f'{split}_set/{uid}/{obj}.nii.gz'

        return zipfile.Path(archive, file)

    @field
    def image(self, i) -> np.ndarray:
        with self._file(i, 'image').open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return np.asarray(image.dataobj).astype(np.int16)

    @field
    def affine(self, i) -> np.ndarray:
        """The 4x4 matrix that gives the image's spatial orientation"""
        with self._file(i, 'image').open('rb') as opened:
            with gzip.GzipFile(fileobj=opened) as nii:
                nii = nibabel.FileHolder(fileobj=nii)
                image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})
                return image.affine

    @field
    def masks(self, i) -> Dict[str, np.ndarray]:
        masks = {}
        for x in range(1, 4):
            with self._file(i, f'annotation_{x}').open('rb') as opened:
                with gzip.GzipFile(fileobj=opened) as nii:
                    nii = nibabel.FileHolder(fileobj=nii)
                    image = nibabel.Nifti1Image.from_file_map({'header': nii, 'image': nii})

                    masks[f'annotation_{x}'] = np.asarray(image.dataobj).astype(np.uint8)

        return masks



import contextlib
import gzip
import json
import tarfile
from pathlib import Path

import nibabel as nb
import numpy as np

from .internals import Dataset, register


@register(
    body_region=('Chest', 'Abdominal', 'Head'),
    link='http://medicaldecathlon.com/',
    modality=('CT', 'CE CT', 'MRI', 'MRI FLAIR', 'MRI T1w', 'MRI t1gd', 'MRI T2w', 'MRI T2', 'MRI ADC'),
    raw_data_size='97.8G',
    task='Image segmentation',
)
class MSD(Dataset):
    """
    MSD is a Medical Segmentaton Decathlon Challenge with 10 tasks.
    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Data can be downloaded here:http://medicaldecathlon.com/
    or here: https://msd-for-monai.s3-us-west-2.amazonaws.com/
    or here: https://drive.google.com/drive/folders/1HqEgzS8BV2c7xYNrZdEAnrHk7osJJ--2/
    Then, the folder with raw downloaded data should contain tar archive with data and masks
    (`Task03_Liver.tar`).
    """

    @property
    def ids(self):
        ids_all = []
        for folder in self.root.glob('*'):
            if folder.name.endswith('.tar'):
                ids_folder = ids_from_tar(folder)
            else:
                ids_folder = ids_from_folder(folder)
            ids_all.extend(ids_folder)
        return tuple(ids_all)

    def train_test(self, i) -> str:
        fold = 'train' if 'train' in i else 'test'
        return fold

    def task(self, i) -> str:
        return NAME_TO_TASK[i.split('_')[1]]

    def _relative(self, i):
        name = i.removeprefix('train_').removeprefix('test_')
        return Path(self.task(i)), Path('imagesTr' if 'train' in i else 'imagesTs') / f'{name}.nii.gz'

    def image(self, i):
        with open_nii_gz(self.root, self._relative(i)) as (file, unpacked):
            if unpacked:
                return np.int16(nb.load(file).get_fdata())
            else:
                with gzip.GzipFile(fileobj=file) as nii_gz:
                    nii = nb.FileHolder(fileobj=nii_gz)
                    return np.int16(nb.Nifti1Image.from_file_map({'header': nii, 'image': nii}).get_fdata())

    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz(self.root, self._relative(i)) as (file, unpacked):
            if unpacked:
                return nb.load(file).affine
            else:
                with gzip.GzipFile(fileobj=file) as nii_gz:
                    nii = nb.FileHolder(fileobj=nii_gz)
                    return nb.Nifti1Image.from_file_map({'header': nii, 'image': nii}).affine

    def image_modality(self, i):
        task = self.task(i)
        if (self.root / task).is_dir():
            with open(self.root / task / 'dataset.json', 'r') as file:
                return json.loads(file.read())['modality']

        with tarfile.open(self.root / f'{task}.tar') as tf:
            member = tf.getmember(f'{task}/dataset.json')
            file = tf.extractfile(member)
            return json.loads(file.read())['modality']

    def segmentation_labels(self, i):
        """Returns segmentation labels for the task"""
        task = self.task(i)
        if (self.root / task).is_dir():
            with open(self.root / task / 'dataset.json', 'r') as file:
                return json.loads(file.read())['labels']

        with tarfile.open(self.root / f'{task}.tar') as tf:
            member = tf.getmember(f'{task}/dataset.json')
            file = tf.extractfile(member)
            return json.loads(file.read())['labels']

    def mask(self, i):
        task, relative = self._relative(i)
        if 'imagesTs' not in str(relative):
            with open_nii_gz(self.root, (task, str(relative).replace('images', 'labels'))) as (file, unpacked):
                if unpacked:
                    return np.uint8(nb.load(file).get_fdata())
                else:
                    with gzip.GzipFile(fileobj=file) as nii_gz:
                        nii = nb.FileHolder(fileobj=nii_gz)
                        return np.uint8(nb.Nifti1Image.from_file_map({'header': nii, 'image': nii}).get_fdata())


TASK_TO_NAME: dict = {
    'Task01_BrainTumour': 'BRATS',
    'Task02_Heart': 'la',
    'Task03_Liver': 'liver',
    'Task04_Hippocampus': 'hippocampus',
    'Task05_Prostate': 'prostate',
    'Task06_Lung': 'lung',
    'Task07_Pancreas': 'pancreas',
    'Task08_HepaticVessel': 'hepaticvessel',
    'Task09_Spleen': 'spleen',
    'Task10_Colon': 'colon',
}

NAME_TO_TASK = dict(zip(TASK_TO_NAME.values(), TASK_TO_NAME.keys()))


@contextlib.contextmanager
def open_nii_gz(path, nii_gz_path):
    """Opens a .nii.gz file from inside a .tar archive.

    Parameters:
    - path: path to the .tar archive or folder
    - nii_gz_path: path to the .nii.gz file inside the .tar archive.

    Yields:
    - nibabel.Nifti1Image object.
    """
    task, relative = nii_gz_path
    if (path / task / relative).exists():
        yield path / task / relative, True
    else:
        with tarfile.open(path / f'{task}.tar', 'r') as tar:
            yield tar.extractfile(str(task / relative)), False


def get_id(filename: Path):
    fold = 'test' if 'imagesTs' in str(filename) else 'train'
    name = filename.name.removesuffix('.nii.gz')
    return '_'.join([fold, name])


def ids_from_tar(tar_folder):
    ids = []
    with tarfile.open(tar_folder, 'r') as tf:
        for file in tf.getmembers():
            filename = Path(file.name)
            if not filename.name.startswith('._') and filename.suffix == '.gz' and 'images' in filename.parent.name:
                ids.append(get_id(filename))
    return sorted(ids)


def ids_from_folder(folder):
    ids = []
    for filename in folder.rglob('*.nii.gz'):
        if not filename.name.startswith('._') and filename.suffix == '.gz' and 'images' in filename.parent.name:
            ids.append(get_id(filename))
    return sorted(ids)



import codecs
import json
import warnings
from functools import cached_property
from pathlib import Path

import numpy as np
import pydicom
from dicom_csv import (
    get_common_tag,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    get_tag,
    order_series,
    stack_images,
)
from dicom_csv.exceptions import TagMissingError
from tqdm.auto import tqdm

from ..internals import Dataset, field, register
from ..utils import get_series_date
from .nodules import get_nodules


@register(
    body_region='Thorax',
    modality='CT',
    task='Lung Cancer Detection',
    link='https://mosmed.ai/en/datasets/mosmeddata-kt-s-priznakami-raka-legkogo-tip-viii/',
    prep_data_size='103G',
    raw_data_size='187G',
)
class MoscowCancer500(Dataset):
    """
    The Moscow Radiology Cancer-500 dataset.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded files.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Download links:
    https://mosmed.ai/en/datasets/mosmeddata-kt-s-priznakami-raka-legkogo-tip-viii/
    After pressing the `download` button you will have to provide an email address to which further instructions
    will be sent.

    Examples
    --------
    >>> # Place the downloaded files in any folder and pass the path to the constructor:
    >>> ds = MoscowCancer500(root='/path/to/files/root')
    >>> print(len(ds.ids))
    # 979
    >>> print(ds.image(ds.ids[0]).shape)
    # (512, 512, 67)
    """

    @cached_property
    def _mapping(self):
        path = self.root / 'series-to-files.json'
        if not path.exists():
            mapping = {}
            for file in tqdm(
                self.root.rglob('*'), total=sum(1 for _ in self.root.rglob('*')), desc='Analyzing folder structure'
            ):
                if file.is_dir():
                    continue

                series = pydicom.dcmread(file, specific_tags=[(0x0020, 0x000E)]).SeriesInstanceUID
                mapping[series].append(str(file.relative_to(self.root)))

            with open(path, 'w') as file:
                json.dump(mapping, file)
            return mapping

        with open(path) as file:
            return json.load(file)

    @property
    def ids(self):
        # this id has an undefined image orientation
        ignore = {'1.2.643.5.1.13.13.12.2.77.8252.604378326291403.583548115656123.'}
        return tuple(sorted(set(self._mapping) - ignore))

    def _series(self, i):
        series = [pydicom.dcmread(Path(self.root, 'dicom', f)) for f in self._mapping[i]]
        series = order_series(series, decreasing=False)
        return series

    @field
    def image(self, i):
        x = stack_images(self._series(i), -1).astype(np.int16)
        # DICOM specifies that the first 2 axes are (y, x). let's fix that
        return np.moveaxis(x, 0, 1)

    @field
    def study_uid(self, i):
        return get_common_tag(self._series(i), 'StudyInstanceUID')

    @field
    def series_uid(self, i):
        return get_common_tag(self._series(i), 'SeriesInstanceUID')

    @field
    def sop_uids(self, i):
        return [str(get_tag(i, 'SOPInstanceUID')) for i in self._series(i)]

    @field
    def pixel_spacing(self, i):
        return get_pixel_spacing(self._series(i)).tolist()

    @field
    def slice_locations(self, i):
        return get_slice_locations(self._series(i))

    @field
    def orientation_matrix(self, i):
        return get_orientation_matrix(self._series(i))

    @field
    def instance_numbers(self, i):
        try:
            instance_numbers = [int(get_tag(i, 'InstanceNumber')) for i in self._series(i)]
            if not _is_monotonic(instance_numbers):
                warnings.warn('Ordered series has non-monotonic instance numbers.')

            return instance_numbers
        except TagMissingError:
            pass

    @field
    def conv_kernel(self, i):
        return get_common_tag(self._series(i), 'ConvolutionKernel', default=None)

    @field
    def kvp(self, i):
        return get_common_tag(self._series(i), 'KVP', default=None)

    @field
    def patient_id(self, i):
        return get_common_tag(self._series(i), 'PatientID', default=None)

    @field
    def study_date(self, i):
        return get_series_date(self._series(i))

    @field
    def accession_number(self, i):
        return get_common_tag(self._series(i), 'AccessionNumber', default=None)

    @field
    def nodules(self, i):
        folders = {Path(f).parent.name for f in self._mapping[i]}
        if len(folders) != 1:
            # can't determine protocol filename
            return

        (filename,) = folders
        protocol = json.load(codecs.open(str(self.root / 'protocols' / f'{filename}.json'), 'r', 'utf-8-sig'))

        series_number = get_common_tag(self._series(i), 'SeriesNumber')
        try:
            return get_nodules(protocol, series_number, self.slice_locations(i))
        except ValueError:
            pass


def _is_monotonic(sequence):
    sequence = list(sequence)
    return sequence == sorted(sequence) or sequence == sorted(sequence)[::-1]



from functools import cached_property

import nibabel as nb
import numpy as np
import pandas as pd

from ..internals import Dataset, licenses, register
from .data_classes import AcquisitionInfo, ClinicalInfo


@register(
    body_region='Head',
    license=licenses.CC_BY_40,
    link='https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70225642',
    modality=('FLAIR', 'MRI T1', 'MRI T1GD', 'MRI T2', 'DSC MRI', 'DTI MRI'),
    prep_data_size='70G',
    raw_data_size='69G',
    task='Segmentation',
)
class UPENN_GBM(Dataset):
    """
    Multi-parametric magnetic resonance imaging (mpMRI) scans for de novo Glioblastoma
      (GBM) patients from the University of Pennsylvania Health System (UPENN-GBM).
    Dataset contains 630 patients.

    All samples are registered to a common atlas (SRI)
        using a uniform preprocessing and the segmentation are aligned with them.


    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70225642
    Download to the root folder nifti images and metadata. Organise folder as folows:


    <...>/<UPENN-root>/NIfTI-files/images_segm/UPENN-GBM-00054_11_segm.nii.gz
    <...>/<UPENN-root>/NIfTI-files/...

    <...>/<UPENN-root>/UPENN-GBM_clinical_info_v1.0.csv
    <...>/<UPENN-root>/UPENN-GBM_acquisition.csv


    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = UPENN_GBM(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 671
    >>> print(ds.image(ds.ids[215]).shape)
    # (4, 240, 240, 155)
    >>> print(d.acqusition_info(d.ids[215]).manufacturer)
    # SIEMENS

    References
    ----------
    .. [1] Bakas, S., Sako, C., Akbari, H., Bilello, M., Sotiras, A., Shukla, G., Rudie,
      J. D., Flores Santamaria, N., Fathi Kazerooni, A., Pati, S., Rathore, S.,
    Mamourian, E., Ha, S. M., Parker, W., Doshi, J., Baid, U., Bergman, M., Binder, Z. A., Verma, R., … Davatzikos,
    C. (2021). Multi-parametric magnetic resonance imaging (mpMRI) scans for de novo
    Glioblastoma (GBM) patients from the University of Pennsylvania Health System (UPENN-GBM)
    (Version 2) [Data set]. The Cancer Imaging Archive.
    https://doi.org/10.7937/TCIA.709X-DN49

    """

    @property
    def ids(self):
        ids = [x.name for x in (self.root / 'NIfTI-files/images_structural').iterdir()]
        return tuple(sorted(ids))

    @property
    def modalities(self):
        return ['T1', 'T1GD', 'T2', 'FLAIR']

    @property
    def dsc_modalities(self):
        return ['', 'ap-rCBV', 'PH', 'PSR']

    @property
    def dti_modalities(self):
        return ['AD', 'FA', 'RD', 'TR']

    def _mask_path(self, i):
        p1 = self.root / 'NIfTI-files/images_segm'
        p2 = self.root / 'NIfTI-files/automated_segm'
        p1 = list(p1.glob(i + '*'))
        p2 = list(p2.glob(i + '*'))
        return p1[0] if p1 else p2[0] if p2 else None

    def mask(self, i):
        path = self._mask_path(i)
        if not path:
            return None
        return np.asarray(nb.load(path).get_fdata())

    def is_mask_automated(self, i):
        path = self._mask_path(i)
        if path is None:
            return None
        return path.parent.name == 'automated_segm'

    def image(self, i):
        path = self.root / f'NIfTI-files/images_structural/{i}'
        image_pathes = [path / f'{i}_{mod}.nii.gz' for mod in self.modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_unstripped(self, i):
        path = self.root / f'NIfTI-files/images_structural_unstripped/{i}'
        image_pathes = [path / f'{i}_{mod}_unstripped.nii.gz' for mod in self.modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_DTI(self, i):
        path = self.root / f'NIfTI-files/images_DTI/{i}'
        if not path.exists():
            return None
        image_pathes = [path / f'{i}_DTI_{mod}.nii.gz' for mod in self.dti_modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return np.stack(images)

    def image_DSC(self, i):
        path = self.root / f'NIfTI-files/images_DSC/{i}'
        if not path.exists():
            return None
        image_pathes = [path / (f'{i}_DSC_{mod}.nii.gz' if mod else f'{i}_DSC.nii.gz') for mod in self.dsc_modalities]
        images = [np.asarray(nb.load(p).dataobj) for p in image_pathes]
        return images

    @cached_property
    def _clinical_info(self):
        return pd.read_csv(self.root / 'UPENN-GBM_clinical_info_v1.0.csv')

    @cached_property
    def _acqusition_info(self):
        return pd.read_csv(self.root / 'UPENN-GBM_acquisition.csv')

    def clinical_info(self, i):
        row = self._clinical_info[self._clinical_info.ID == i]
        return ClinicalInfo(*row.iloc[0, 1:])

    def acqusition_info(self, i):
        row = self._acqusition_info[self._acqusition_info.ID == i]
        return AcquisitionInfo(*row.iloc[0, 1:])

    def subject_id(self, i):
        return i.split('_')[0]

    def affine(self, i):
        return np.array([[-1.0, 0.0, 0.0, -0.0], [0.0, -1.0, 0.0, 239.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0]])

    def spacing(self, i):
        return (1, 1, 1)



from enum import Enum
from typing import NamedTuple, Optional, Sequence

import numpy as np


class Calcification(Enum):
    Popcorn, Laminated, Solid, NonCentral, Central, Absent = 1, 2, 3, 4, 5, 6


class InternalStructure(Enum):
    SoftTissue, Fluid, Fat, Air = 1, 2, 3, 4


class Lobulation(Enum):
    NoLobulation, NearlyNoLobulation, MediumLobulation, NearMarkedLobulation, MarkedLobulation = 1, 2, 3, 4, 5


class Malignancy(Enum):
    HighlyUnlikely, ModeratelyUnlikely, Indeterminate, ModeratelySuspicious, HighlySuspicious = 1, 2, 3, 4, 5


class Sphericity(Enum):
    Linear, OvoidLinear, Ovoid, OvoidRound, Round = 1, 2, 3, 4, 5


class Spiculation(Enum):
    NoSpiculation, NearlyNoSpiculation, MediumSpiculation, NearMarkedSpiculation, MarkedSpiculation = 1, 2, 3, 4, 5


class Subtlety(Enum):
    ExtremelySubtle, ModeratelySubtle, FairlySubtle, ModeratelyObvious, Obvious = 1, 2, 3, 4, 5


class Texture(Enum):
    NonSolidGGO, NonSolidMixed, PartSolidMixed, SolidMixed, Solid = 1, 2, 3, 4, 5


class LIDCNodule(NamedTuple):
    center_voxel: Sequence[float]
    bbox: np.ndarray
    diameter_mm: float
    surface_area_mm2: float
    volume_mm3: float
    calcification: Optional[Calcification] = None
    internal_structure: Optional[InternalStructure] = None
    lobulation: Optional[Lobulation] = None
    malignancy: Optional[Malignancy] = None
    sphericity: Optional[Sphericity] = None
    spiculation: Optional[Spiculation] = None
    subtlety: Optional[Subtlety] = None
    texture: Optional[Texture] = None



import pickle

import numpy as np
import pytest

from amid.internals import gather_datasets


MAPPING = gather_datasets()
DATASETS = [x[0] for x in MAPPING.values()]
NAMES = list(MAPPING)


@pytest.mark.raw
@pytest.mark.parametrize('cls', DATASETS, ids=NAMES)
def test_ids_availability(cls):
    assert len(cls().ids) > 0


@pytest.mark.raw
@pytest.mark.parametrize('cls', DATASETS, ids=NAMES)
def test_pickleable(cls):
    raw = cls()[0]
    cached = cls()
    fields = dir(raw)

    for ds in raw, cached:
        loader = ds._compile(fields)
        pickle.dumps(loader)

    f = cached._compile('ids')
    raw = pickle.dumps(f)
    g = pickle.loads(raw)
    assert f() == g()


# @pytest.mark.raw
# @pytest.mark.parametrize('cls', ROOT_MAPPING, ids=[cls.__name__ for cls in ROOT_MAPPING])
# def test_cache_consistency(cls):
#     raw = cls(root=ROOT_MAPPING[cls])
#     cached = raw.cached()
#     fields = {x.name for x in raw._container.outputs} - {'ids', 'id', 'cached'}
#
#     ids = raw.ids
#     assert ids == cached.ids
#     for i in ids:
#         for field in fields:
#             compare(getattr(raw, field)(i), getattr(cached, field)(i))


# TODO: find a package for this
def compare(x, y):
    assert type(x) == type(y)
    if isinstance(x, (str, int, float, bytes)):
        assert x == y
    elif isinstance(x, (np.ndarray, np.generic)):
        np.testing.assert_allclose(x, y)
    elif isinstance(x, (list, tuple)):
        list(map(compare, x, y))
    else:
        raise TypeError(type(x))



from typing import NamedTuple


class License(NamedTuple):
    name: str
    url: str


CC0_10 = License(name='CC0 1.0', url='https://creativecommons.org/publicdomain/zero/1.0/')
CC_BY_30 = License(
    name='CC BY 3.0',
    url='https://creativecommons.org/licenses/by/3.0/',
)
CC_BY_40 = License(
    name='CC BY 4.0',
    url='https://creativecommons.org/licenses/by/4.0/',
)
CC_BYNC_40 = License(
    name='CC BY-NC 4.0',
    url='https://creativecommons.org/licenses/by-nc/4.0/',
)
CC_BYND_40 = License(
    name='CC BY-ND 4.0',
    url='https://creativecommons.org/licenses/by-nd/4.0/',
)
CC_BYNCND_40 = License(
    name='CC BY-NC-ND 4.0',
    url='https://creativecommons.org/licenses/by-nc-nd/4.0/',
)
CC_BYSA_40 = License(
    name='CC BY-SA 4.0',
    url='https://creativecommons.org/licenses/by-sa/4.0/',
)
CC_BYNCSA_40 = License(
    name='CC BY-NC-SA 4.0',
    url='https://creativecommons.org/licenses/by-nc-sa/4.0/',
)

PhysioNet_RHD_150 = License(
    name='PhysioNet Restricted Health Data License 1.5.0',
    url='https://www.physionet.org/about/licenses/physionet-restricted-health-data-license-150/',
)

StanfordDSResearch = License(
    name='Stanford University Dataset Research Use Agreement',
    url='https://stanfordaimi.azurewebsites.net/datasets/e8ca74dc-8dd4-4340-815a-60b41f6cb2aa',  # TODO: separate link
)



import json
import warnings
from functools import cached_property
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import pydicom
from dicom_csv import (
    drop_duplicated_instances,
    drop_duplicated_slices,
    expand_volumetric,
    get_orientation_matrix,
    get_pixel_spacing,
    get_slice_locations,
    join_tree,
    order_series,
    stack_images,
)

from .internals import Dataset, field, licenses, register


@register(
    body_region='Thorax',
    license=licenses.CC_BY_30,
    link='https://wiki.cancerimagingarchive.net/display/Public/NSCLC-Radiomics',
    modality='CT',
    prep_data_size='13G',
    raw_data_size='34G',
    task='Tumor Segmentation',
)
class NSCLC(Dataset):
    """

        NSCLC-Radiomics is a public cell lung cancer segmentation dataset with 422 patients.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.


    Notes
    -----
    Follow the download instructions at https://wiki.cancerimagingarchive.net/display/Public/NSCLC-Radiomics

    The folder with downloaded data should contain two paths

    The folder should have this structure:
        <...>/<NSCLC-root>/NSCLC-Radiomics/LUNG1-XXX


    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = NSCLC(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
     422
    >>> print(ds.image(ds.ids[0]).shape)
     (512, 512, 134)
    >>> print(ds.mask(ds.ids[80]).shape)
     (512, 512, 108)

    References
    ----------
    """

    # FIXME: move to filtering via `ignore_errors=True` + filtering via Filter (if needed)
    _INVALID_PATIENT_IDS = (
        # no dicom with cancer segmentation
        'LUNG1-128',
        'LUNG1-412',
        # image.shape != cancer.shape
        'LUNG1-194',
        'LUNG1-095',
        'LUNG1-085',
        'LUNG1-014',
        'LUNG1-021',
    )

    @cached_property
    def _joined(self):
        joined_path = self.root / 'joined.csv'
        if joined_path.exists():
            return pd.read_csv(joined_path)
        joined = join_tree(self.root / 'NSCLC-Radiomics', verbose=1)
        joined = joined[[x.endswith('.dcm') for x in joined.FileName]]
        joined.to_csv(joined_path)
        return joined

    @property
    def ids(self):
        uid = self._joined.groupby('SeriesInstanceUID').apply(len)
        return tuple(uid[uid > 1].keys())

    def _sub(self, i):
        return self._joined[self._joined.SeriesInstanceUID == i]

    def _series(self, i):
        series = [
            pydicom.dcmread(self.root / 'NSCLC-Radiomics' / file.PathToFolder / file.FileName)
            for _, file in self._sub(i).iterrows()
        ]
        series = expand_volumetric(series)
        series = drop_duplicated_instances(series)

        if True:  # drop_dupl_slices
            _original_num_slices = len(series)
            series = drop_duplicated_slices(series)
            if len(series) < _original_num_slices:
                warnings.warn(f'Dropped duplicated slices for series {series[0]["StudyInstanceUID"]}.')

        return order_series(series, decreasing=False)

    @field
    def image(self, i) -> np.ndarray:
        image = stack_images(self._series(i), -1).astype(np.int16).transpose(1, 0, 2)
        return image

    @field
    def image_meta(self, i) -> dict:
        metas = [list(dict(s).values()) for s in self._series(i)]
        result = {}
        for meta_ in metas:
            for element in meta_:
                if element.keyword in ['PixelData']:
                    continue
                if element.keyword not in result:
                    result[element.keyword] = [element.value]
                elif result[element.keyword][-1] != element.value:
                    result[element.keyword].append(element.value)
        # turn elements that are the same across the series back from array
        result = {k: v[0] if len(v) == 1 else v for k, v in result.items()}
        return result

    @field
    def sex(self, i) -> str:
        """Sex of the patient."""
        return self._sub(i)['PatientSex'].iloc[1]

    @field
    def age(self, i) -> Union[int, None]:
        """Age of the patient, dataset contains 97 patients with unknown Age."""
        age = self._sub(i)['PatientAge'].iloc[1]
        if isinstance(age, str):
            return int(age.removesuffix('Y'))
        return age

    def _study_id(self, i):
        study_ids = self._joined[self._joined.SeriesInstanceUID == i].StudyInstanceUID.unique()
        assert len(study_ids) == 1
        # series_id_to_study
        return study_ids[0]

    @field
    def spacing(self, i) -> np.ndarray:
        pixel_spacing = get_pixel_spacing(self._series(i)).tolist()
        slice_locations = get_slice_locations(self._series(i))
        diffs, counts = np.unique(np.round(np.diff(slice_locations), decimals=5), return_counts=True)
        spacing = np.float32([pixel_spacing[1], pixel_spacing[0], diffs[np.argsort(counts)[-1]]])
        return spacing

    @field
    def mask(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('GTV-1', None)

    @field
    def lung_left(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Lung-Left', None)

    @field
    def lung_right(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Lung-Right', None)

    @field
    def lungs_total(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Lungs-Total', None)

    @field
    def heart(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Heart', None)

    @field
    def esophagus(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Esophagus', None)

    @field
    def spinal_cord(self, i) -> np.ndarray:
        return self._extract_segment_masks(i).get('Spinal-Cord', None)

    def _extract_segment_masks(self, i):
        folders = self._joined[self._joined.SeriesInstanceUID == i].PathToFolder.unique()
        assert len(folders) == 1, i
        patient_id = self._joined[self._joined.SeriesInstanceUID == i].PatientID.unique()[0]
        if patient_id in self._INVALID_PATIENT_IDS:
            return {}

        annotation_path = self.root / 'NSCLC-Radiomics' / Path(folders[0]).parent

        found_markup = None
        for p in annotation_path.glob('*.json'):
            with open(p, 'r') as f:
                data = json.load(f)
            if 'Segmentation' in data['Total'][-1]:
                assert found_markup is None, annotation_path
                found_markup = data

        if found_markup is None:
            raise FileNotFoundError

        dicom_pathes = list((annotation_path / found_markup['SeriesUID']).glob('*'))
        assert len(dicom_pathes) == 1, annotation_path
        cancer_dicom = pydicom.dcmread(dicom_pathes[0])
        assert np.allclose(get_orientation_matrix(self._series(i)), get_cancer_orientation_matrix(cancer_dicom)), i
        mask = np.moveaxis(cancer_dicom.pixel_array, 0, -1).astype(bool).transpose(1, 0, 2)
        mask_slice_locations = get_mask_slice_locations(cancer_dicom)
        slice_locations = get_slice_locations(self._series(i))
        image = stack_images(self._series(i), -1).transpose(1, 0, 2)
        segments = [x.SegmentDescription for x in cancer_dicom.SegmentSequence]
        assert len(mask_slice_locations) == len(slice_locations) * len(segments), i

        all_masks = {}
        for n, seg in enumerate(segments):
            mask_subslice = slice(
                len(slice_locations) * n, len(slice_locations) * (n + 1) if n + 1 != len(segments) else None
            )
            sub_mask = mask[:, :, mask_subslice]
            sub_mask_slice_locations = mask_slice_locations[mask_subslice]

            assert sub_mask.shape == image.shape, i
            if np.allclose(sub_mask_slice_locations, slice_locations, atol=0.01):
                pass
            elif np.allclose(sub_mask_slice_locations, slice_locations[::-1], atol=0.01):
                sub_mask = sub_mask[..., ::-1]
            else:
                raise AssertionError(i)
            all_masks[seg] = sub_mask
        return all_masks


def get_cancer_orientation_matrix(cancer_dicom):
    row, col = np.array(
        [
            float(x)
            for x in cancer_dicom.SharedFunctionalGroupsSequence[0].PlaneOrientationSequence[0].ImageOrientationPatient
        ]
    ).reshape(2, 3)
    return np.stack([row, col, np.cross(row, col)])


def get_mask_slice_locations(cancer_dicom):
    om = get_cancer_orientation_matrix(cancer_dicom)
    image_position_patient = np.stack(
        [
            list(map(float, frame.PlanePositionSequence[0].ImagePositionPatient))
            for frame in cancer_dicom.PerFrameFunctionalGroupsSequence
        ]
    )
    return list(image_position_patient @ om[-1])



from pathlib import Path

import nibabel as nb

from ..internals import Dataset, licenses, register


@register(
    body_region='Head',
    license=licenses.CC_BY_30,
    link='https://github.com/muschellij2/open_ms_data?tab=readme-ov-file',
    modality='MRI',
    prep_data_size='18G',
    raw_data_size='5.9G',
    task='Anomaly segmentation',
)
class MSLUB(Dataset):
    @property
    def ids(self):
        result = set()
        for file in self.root.glob('**/*.gz'):
            if ('raw' not in str(file)) or ('gt' in str(file)):
                continue
            patient = file.parent.name
            plane = file.parent.parent.parent.name
            ind = f'{plane}-{patient}'
            if 'longitudinal' in str(file):
                filename = file.name
                study_number = filename.split('_')[0]
                ind = f'{ind}-{study_number}'
            result.add(ind)
        return list(result)

    def _file(self, i):
        plane = i.split('-')[0]
        patient = i.split('-')[1]
        path = self.root / plane / 'raw' / patient
        if 'longitudinal' in i:
            study_number = i.split('-')[2]
            return path / study_number
        return path

    def image(self, i):
        file = self._file(i)
        if 'longitudinal' in str(file):
            study_number = file.stem
            file_name = file.parent / f'{study_number}_FLAIR.nii.gz'
        else:
            file_name = file / 'FLAIR.nii.gz'
        image = nb.load(file_name).get_fdata()
        return image

    def mask(self, i):
        file = self._file(i)
        if 'longitudinal' in str(file):
            file_name = file.parent / 'gt.nii.gz'
        else:
            file_name = file / 'consensus_gt.nii.gz'
        image = nb.load(file_name).get_fdata()
        return image

    def patient(self, i):
        file = self._file(i)
        if 'longitudinal' in str(file):
            return Path(file).parent.name
        else:
            return Path(file).name

    def affine(self, i):
        file = self._file(i)
        if 'longitudinal' in str(file):
            study_number = file.stem
            file_name = file.parent / f'{study_number}_FLAIR.nii.gz'
        else:
            file_name = file / 'FLAIR.nii.gz'
        return nb.load(file_name).affine



import contextlib
import gzip
import zipfile
from pathlib import Path
from zipfile import ZipFile

import nibabel as nb
import numpy as np

from .internals import Dataset, field, licenses, register


@register(
    body_region='Chest',
    license=licenses.CC0_10,
    link='http://medicalsegmentation.com/covid19/',
    modality='CT',
    prep_data_size='300M',
    raw_data_size='310M',
    task='COVID-19 segmentation',
)
class Medseg9(Dataset):
    """

    Medseg9 is a public COVID-19 CT segmentation dataset with 9 annotated images.

    Parameters
    ----------
    root : str, Path, optional
        path to the folder containing the raw downloaded archives.
        If not provided, the cache is assumed to be already populated.

    Notes
    -----
    Data can be downloaded here: http://medicalsegmentation.com/covid19/.

    Then, the folder with raw downloaded data should contain three zip archives with data and masks
    (`rp_im.zip`, `rp_lung_msk.zip`, `rp_msk.zip`).

    Examples
    --------
    >>> # Place the downloaded archives in any folder and pass the path to the constructor:
    >>> ds = Medseg9(root='/path/to/downloaded/data/folder/')
    >>> print(len(ds.ids))
    # 9
    >>> print(ds.image(ds.ids[0]).shape)
    # (630, 630, 45)
    >>> print(ds.covid(ds.ids[0]).shape)
    # (630, 630, 45)

    """

    @property
    def ids(self):
        result = set()

        with ZipFile(self.root / 'rp_msk.zip') as zf:
            for zipinfo in zf.infolist():
                if zipinfo.is_dir():
                    continue
                file_stem = Path(zipinfo.filename).stem
                result.add('medseg9_' + file_stem.split('.nii')[0])

        return tuple(sorted(result))

    @staticmethod
    def _filename(i):
        num_id = i.split('_')[-1]
        return f'{num_id}.nii.gz'

    def _file(self, i):
        return zipfile.Path(self.root / 'rp_im.zip', f'rp_im/{self._filename(i)}')

    @field
    def image(self, i):
        with open_nii_gz_file(self._file(i)) as nii_image:
            # most CT/MRI scans are integer-valued, this will help us improve compression rates
            return np.int16(nii_image.get_fdata())

    @field
    def affine(self, i):
        """The 4x4 matrix that gives the image's spatial orientation."""
        with open_nii_gz_file(self._file(i)) as nii_image:
            return nii_image.affine

    @field
    def lungs(self, i):
        mask_file = zipfile.Path(self.root / 'rp_lung_msk.zip', f'rp_lung_msk/{self._filename(i)}')
        with open_nii_gz_file(mask_file) as nii_image:
            return np.bool_(nii_image.get_fdata())

    @field
    def covid(self, i):
        """
        int16 mask.
        0 - normal, 1 - ground-glass opacities (матовое стекло), 2 - consolidation (консолидация).
        """
        mask_file = zipfile.Path(self.root / 'rp_msk.zip', f'rp_msk/{self._filename(i)}')
        with open_nii_gz_file(mask_file) as nii_image:
            # most CT/MRI scans are integer-valued, this will help us improve compression rates
            return np.uint8(nii_image.get_fdata())


# TODO: sync with amid.utils
@contextlib.contextmanager
def open_nii_gz_file(file):
    with file.open('rb') as opened:
        with gzip.GzipFile(fileobj=opened) as nii:
            nii = nb.FileHolder(fileobj=nii)
            yield nb.Nifti1Image.from_file_map({'header': nii, 'image': nii})
